/*
Copyright (C) 1995 John Leveling

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

This is Version 0.8.2 alpha of WinPCS, a
text-window extension for use with SCM.

The author can be reached at olga@informatik.uni-oldenburg.de or
J. Leveling, Sonnenstr. 32, 26123 Oldenburg, Germany
*/

#include "win.h"
#ifdef WINEXT
	#include "winext.h"
#endif

/* make-window */
/* no known bugs */
/* inserting and scrolling are initially enabled */
SCM p_mkwin(SCM args)
{
  int nargs=ilength(args);
  SCM win, label, borderp, winattrs;
  WINDOW *nwin;
  ASSERT (WINMODEP, args, e_invmode, s_mkwin);
  ASSERT (nargs<=2, args, WNA, s_mkwin);
  ASSERT (NULLP(args) || (CAR(args)==BOOL_F) || STRINGP(CAR(args)),
	  CAR(args), e_invarg, s_mkwin);
  ASSERT (NULLP(args) || NULLP(CDR(args)) || (CAR(CDR(args))==BOOL_F) ||
	  (CAR(CDR(args))==BOOL_T), CAR(CDR(args)), e_invarg, s_mkwin);
  nwin=newwin(0,0,0,0);
  win=makewin(nwin);
  idlok(WIN(win), TRUE);
  scrollok(WIN(win), TRUE);
  keypad(WIN(win), TRUE);
  wsetscrreg(WIN(win), 0, MAXY(win));
  borderp=BOOL_F;
  if (nargs==0 || (CAR(args)==BOOL_F)) 
     label=CAR(intern(CHARS(""),(sizet)0));
  else 
     {
     label=CAR(args);
     label=CAR(intern(CHARS(label),(sizet)LENGTH(label)));
     }
  if (nargs==2)
     {
     borderp=CAR(CDR(args));
     if (borderp!=BOOL_T) borderp=BOOL_F;
     }
  winattrs=cons(win,				   /* port */
	        cons(borderp,			   /* window-bordered? */
		     cons(label,		   /* window's title */
			  cons(MAKINUM(TEXTATTR), /* attributes */
			       cons(MAKINUM(BORDERATTR), /* borderattrs */
				    EOL)))));
  *loc_winlist=cons(winattrs, *loc_winlist);
  return win;
}

/* window-clear */
/* no known bugs */
/* window contents can be scrolled either by wclear or wscrl */
/* which is faster depends on hardware */
SCM p_winclear (SCM win)
{
#ifdef CLEAR_BY_SCROLL
  short lines;
#endif
  ASSERT (WINMODEP, win, e_invmode, s_winclear);
  SETCONSOLE;
  ASSERT(WINDOWP(win), win, e_invwinarg, s_winclear);
/* sometimes clearing the window by scrolling its contents can be faster */
/* this is hardware-dependant */
#ifdef CLEAR_BY_SCROLL
  lines=MAXY(win);
  wscrl(WIN(win),lines);
  CURY(win)=0;
  CURX(win)=0;
#else
  wclear(WIN(win));
#endif
  borderwin(win, 1);
  wrefresh(WIN(win));
  return UNSPECIFIED;
}

/* window-delete */
/* no known bugs */
/* closes port associated with win */
/* does not allow to delete console */
SCM p_windelete(SCM win)
{
  ASSERT (WINMODEP, win, e_invmode, s_windelete);
  SETCONSOLE;
  ASSERT(WINDOWP(win), win, e_invwinarg, s_windelete);
  borderwin(win, 0);
  wclear(WIN(win));
  wrefresh(WIN(win));
  if (win!=console) close_port(win);
  return UNSPECIFIED;
}

/* window-get-attribute */
/* no known bugs */

SCM p_wingattr(SCM win, SCM attrtype)
{
  SCM wattr;
  ASSERT (WINMODEP, win, e_invmode, s_wingattr);
  SETCONSOLE;
  ASSERT(WINDOWP(win), win, e_invwinarg, s_wingattr);
  ASSERT((attrtype==tattrsym)||(attrtype==battrsym), 
         attrtype, e_invarg, s_wingattr);
  wattr=wingattr(win);
  if (attrtype==tattrsym) return (CAR(CDR(CDR(CDR(wattr)))));
  if ((attrtype==battrsym)&&(CAR(CDR(wattr))==BOOL_T)) 
     return (CAR(CDR(CDR(CDR(CDR(wattr))))));
  else 
     return MAKINUM(-1);
}

/* window-get-cursor */
/* no known bugs */
SCM p_wingcurs(SCM win)
{
  short y, x;
  ASSERT (WINMODEP, win, e_invmode, s_wingcurs);
  SETCONSOLE;
  ASSERT (WINDOWP(win), win, e_invwinarg, s_wingcurs);
  y=MAKINUM(CURY(win));
  x=MAKINUM(CURX(win));
  return cons (y, x);
}

/* window-get-position */
/* no known bugs */
SCM p_wingpos(SCM win)
{
  short line, column;
  ASSERT (WINMODEP, win, e_invmode, s_wingpos);
  SETCONSOLE;
  ASSERT(WINDOWP(win), win, e_invwinarg, s_wingpos);
  line=MAKINUM(BEGY(win));
  column=MAKINUM(BEGX(win));
  return cons (line, column);
}

/* window-get-size */
/* no known bugs */
SCM p_wingsize(SCM win)
{
  short lines, columns;
  ASSERT (WINMODEP, win, e_invmode, s_wingpos);
  SETCONSOLE;
  ASSERT(WINDOWP(win), win, e_invwinarg, s_wingsize);
  lines=MAKINUM(MAXY(win)+1); 
/* + 1 ??? */
  columns=MAKINUM(MAXX(win)+1);
  return cons(lines, columns);
}

/* window-restore-contents */
/* no known bugs */
SCM p_winrestorec(SCM win, SCM contents)
{
  register linec, columnc;
  short lines, columns;
  ASSERT (WINMODEP, win, e_invmode, s_winrestorec);
  SETCONSOLE;
  ASSERT (WINDOWP(win), win, e_invwinarg, s_winrestorec);
  ASSERT (CONSP(contents), contents, e_invarg, s_winrestorec);
  lines=MAXY(win)-1;
  columns=MAXX(win)-1;
  wclear(WIN(win));
/*CURY(win)=CURX(win)=0;*/
  for (linec=0;linec<=lines;linec++)
      for(columnc=0;columnc<=columns;columnc++)
         {
	 ASSERT (INUMP(CAR(contents)), CAR(contents), 
		 e_invarg, s_winrestorec);
         mvwinsch(WIN(win), linec, columnc, (chtype)INUM(CAR(contents)));
         contents=CDR(contents);
         }
/*  borderwin(win, 1);*/
  /* reset attributes */
  wrefresh(WIN(win));
  return UNSPECIFIED;
}

/* window-save-contents */
/* no known bugs */
SCM p_winsavec(SCM win)
{
  register linec, columnc;
  short lines, columns;
  SCM contents=EOL;
  ASSERT (WINMODEP, win, e_invmode, s_winsavec);
  SETCONSOLE;
  ASSERT (WINDOWP(win), win, e_invwinarg, s_winsavec);
  /*contents=wingattr(win);*/  /* for future restoration of attributes */
  lines=MAXY(win)-1;
  columns=MAXX(win)-1;
  for (linec=lines;linec>=0;linec--)
      for(columnc=columns;columnc>=0;columnc--)
         {
         contents=cons(MAKINUM(mvwinch(WIN(win), linec, columnc)), contents);
         }
  return contents;
}

/* window-scroll-down */
/* no known bugs */
SCM p_winscrldn(SCM win, SCM lines)
{
  ASSERT (WINMODEP, win, e_invmode, s_winscrldn);
  SETCONSOLE;
  ASSERT (WINDOWP(win), win, e_invwinarg, s_winscrldn);
  ASSERT (UNBNDP(lines) || INUMP(lines), lines, e_invarg, s_winscrldn);
  if UNBNDP(lines) lines=MAKINUM(1);
  wscrl(WIN(win),-INUM(lines));
  wrefresh(WIN(win));
  return UNSPECIFIED;
}

/* window-scroll-up */
/* no known bugs */
SCM p_winscrlup(SCM win, SCM lines)
{
  ASSERT (WINMODEP, win, e_invmode, s_winscrlup);
  SETCONSOLE;
  ASSERT (WINDOWP(win), win, e_invwinarg, s_winscrlup);
  ASSERT (UNBNDP(lines) || INUMP(lines), lines, e_invarg, s_winscrlup);
  if UNBNDP(lines) lines=MAKINUM(1);
  wscrl(WIN(win), INUM(lines));
/* This way is more compatible : */
/*  line=CURY(win);
    column=CURX(win);
    CURY(win)=0;
    CURX(win)=0;
    wdeleteln(WIN(win));
    CURY(win)=line;
    CURX(win)=column; */
  wrefresh(WIN(win));
  return UNSPECIFIED;
}

/* window-set-attribute */
/* no known bugs */
SCM p_winsattr(SCM win, SCM attrtype, SCM wattr)
{
  SCM attrs;
  ASSERT (WINMODEP, win, e_invmode, s_winsattr);
  SETCONSOLE;
  ASSERT(WINDOWP(win), win, e_invwinarg, s_winsattr);
  ASSERT((attrtype==tattrsym)||(attrtype==battrsym), 
	 attrtype, e_invarg, s_winsattr);
  ASSERT(NUMBERP(wattr), wattr, e_invarg, s_winsattr);
  if (INUMP(wattr))
     wattr=INUM(wattr);
     else wattr=REALPART(wattr);
  attrs=wingattr(win);
  if (attrtype==tattrsym) 
     CAR(CDR(CDR(CDR(attrs))))=MAKINUM(wattr);
  if (attrtype==battrsym) 
     {
     CAR(CDR(CDR(CDR(CDR(attrs)))))=MAKINUM(wattr);
/*     borderwin(win, 1);*/
     }
  if (attrtype==tattrsym) 
     wattrset(WIN(win), attributes(MAKINUM(wattr)));
  return UNSPECIFIED;
}

/* window-set-cursor! */
/* no known bugs */
SCM p_winscurs(SCM win, SCM cursor_line, SCM cursor_column)
{
  short lines, columns;
  ASSERT (WINMODEP, win, e_invmode, s_winscurs);
  SETCONSOLE;
  ASSERT(WINDOWP(win), win, e_invwinarg, s_winscurs);
  ASSERT(INUMP(cursor_line)||cursor_line==BOOL_F, 
         cursor_line, e_invarg, s_winscurs);
  ASSERT(INUMP(cursor_column)||cursor_column==BOOL_F, 
         cursor_column, e_invarg, s_winscurs);
  if (cursor_line!=BOOL_F)
     {
     lines=MAXY(win);
     if (INUMP(cursor_line))
	cursor_line=INUM(cursor_line);
	else cursor_line=REALPART(cursor_line);
     ASSERT((cursor_line>=0)&&(cursor_line<=lines), 
 /* <= ??? */
	    MAKINUM(cursor_line), e_invlineno, s_winscurs);
     }
  if (cursor_column!=BOOL_F)
     {
     columns=MAXX(win);
     if (INUMP(cursor_column))
	cursor_column=INUM(cursor_column);
	else cursor_column=REALPART(cursor_column);
     ASSERT((cursor_column>=0)&&(cursor_column<=columns), 
	    MAKINUM(cursor_column), e_invcolno, s_winscurs);
     }
  if (cursor_line!=BOOL_F) CURY(win)=cursor_line;
  if (cursor_column!=BOOL_F) CURX(win)=cursor_column;
/* borderwin(win, 1); */
  return UNSPECIFIED;
}

/* window-set-position! */
/* no known bugs */
SCM p_winspos(SCM win, SCM ul_line, SCM ul_column)
{
  short lines,columns;
  ASSERT (WINMODEP, win, e_invmode, s_winspos);
  SETCONSOLE;
  ASSERT(WINDOWP(win), win, e_invwinarg, s_winspos);
  ASSERT(NUMBERP(ul_line)||ul_line==BOOL_F, 
         ul_line, e_invarg, s_winspos);
  ASSERT(NUMBERP(ul_column)||ul_column==BOOL_F, 
         ul_column, e_invarg, s_winspos);
  if (ul_line!=BOOL_F)
     {
     lines=MAXY(win);
     if (INUMP(ul_line)) 
	ul_line=INUM(ul_line);
	else ul_line=REALPART(ul_line);
     ASSERT((ul_line>=0)&&(ul_line<LINES), 
	    MAKINUM(ul_line), e_invlineno, s_winspos);
     }
  if (ul_column!=BOOL_F)
     {
     columns=MAXX(win);
     if (INUMP(ul_column)) 
	ul_column=INUM(ul_column);
	else ul_column=REALPART(ul_column);
     ASSERT((ul_column>=0)&&(ul_column<COLS), 
	    MAKINUM(ul_column), e_invcolno, s_winspos);
     }
  if (ul_line!=BOOL_F) BEGY(win)=ul_line;
  if (ul_column!=BOOL_F) BEGX(win)=ul_column;
/*  borderwin(win, 1);*/
  return UNSPECIFIED;
}

/* window-set-size! */
/* no known bugs */
SCM p_winssize(SCM win, SCM n_lines, SCM n_cols)
{
  short ul_line, ul_column;
  ASSERT (WINMODEP, win, e_invmode, s_winssize);
  SETCONSOLE;
  ASSERT (WINDOWP(win), win, e_invwinarg, s_winssize);
  ASSERT(INUMP(n_lines)||n_lines==BOOL_F, 
         n_lines, e_invarg, s_winssize);
  ASSERT(INUMP(n_cols)||n_cols==BOOL_F, 
         n_cols, e_invarg, s_winssize);
  if (n_lines!=BOOL_F)
     {
     ul_line=BEGY(win);
     if (INUMP(n_lines)) 
	n_lines=INUM(n_lines);
	else n_lines=REALPART(n_lines);
     ASSERT((n_lines>0)&&(n_lines<=LINES), 
	    MAKINUM(n_lines), e_invlineno, s_winssize);
     } 
/* >0 or >= 0 ??? */
  if (n_cols!=BOOL_F)
     {
     ul_column=BEGX(win);
     if (INUMP(n_cols))
	n_cols=INUM(n_cols);
	else n_cols=REALPART(n_cols);
     ASSERT((n_cols>0)&&(n_cols<=COLS), 
	    MAKINUM(n_cols), e_invcolno, s_winssize);
     }
  if (n_lines!=BOOL_F) MAXY(win)=n_lines-1;
  if (n_cols!=BOOL_F) MAXX(win)=n_cols-1;
  wsetscrreg(WIN(win), 0, MAXY(win));
/*  borderwin(win, 1);*/
  return UNSPECIFIED;
}

/* window ? */
/* no known bugs */
SCM p_winp(SCM obj)
{
  ASSERT (WINMODEP, obj, e_invmode, s_winp);
  return ((obj==consolesym) ||(NIMP(obj) && WINP(obj))) ? BOOL_T : BOOL_F;
}


/* enter-winmode */
/* no known bugs */
SCM p_ewinmode()
{
  if (winmodep==1) return BOOL_F;
  winmodep=1;
  wclear(WIN(console));
  wrefresh(WIN(console));
  idlok(WIN(console), TRUE);
  scrollok(WIN(console), TRUE);
  wsetscrreg(WIN(console), 0, MAXY(console));
  cur_errp=console;
  cur_outp=console;
  cur_inp=console;
  return BOOL_T;
}

/* exit-winmode */
/* no known bugs */
SCM p_xwinmode()
{
  if (winmodep==0) return BOOL_F;
  winmodep=0;
  if IMP(*loc_console) return BOOL_F;
  wclear(WIN(console));  
  wrefresh(WIN(console));
  cur_errp=def_errp;
  cur_inp=def_inp;
  cur_outp=def_outp;
  return ERR==endwin() ? BOOL_F : BOOL_T;
}


static iproc subr0s[] = {
	{s_ewinmode, p_ewinmode},
	{s_xwinmode, p_xwinmode},
#ifdef WINEXT
	{s_winghelp, p_winghelp},
#endif
	{0,0}};

static iproc subr1s[] = {
	{s_winp, p_winp},
	{s_windelete, p_windelete},
	{s_winclear, p_winclear},
	{s_wingcurs, p_wingcurs},
	{s_wingsize, p_wingsize},
	{s_wingpos, p_wingpos},
	{s_winsavec, p_winsavec},
#ifdef WINEXT
	{s_winrefresh, p_winrefresh},
#endif
	{0,0}};

static iproc subr2s[] = {
	{s_winrestorec, p_winrestorec},
	{s_wingattr, p_wingattr},
	{0,0}};

static iproc subr3s[] = {
	{s_winscurs, p_winscurs},
	{s_winspos, p_winspos},
	{s_winssize, p_winssize},
	{s_winsattr, p_winsattr},
	{0,0}};

static iproc lsubrs[] = {
	{s_mkwin, p_mkwin},
#ifdef WINEXT
	{s_newwin, p_newwin},
	{s_windisplay, p_windisplay},
#endif
	{0,0}};

static iproc subr2os[] = {
	{s_winscrlup, p_winscrlup},
	{s_winscrldn, p_winscrldn},
#ifdef WINEXT
	{s_winborder, p_winborder},
	{s_wintitle, p_wintitle},
#endif
	{0,0}};


/* initialization */
void init_win()
{
  WINDOW *conwin;
  char s_console[]="console",
       s_tattr[]="text-attributes",
       s_battr[]="border-attributes",
       s_nodsp[]="no-display";
  loc_console=&CDR(sysintern("console", UNDEFINED));
  tc16_window=newptob(&winptob);
  loc_winlist=&CDR(sysintern("*window-list*", EOL));  
  loc_maxconsole=&CDR(sysintern("max-console", EOL));
  consolesym=string2symbol(makfromstr(s_console, strlen(s_console)));
  tattrsym=string2symbol(makfromstr(s_tattr, strlen(s_tattr)));
  battrsym=string2symbol(makfromstr(s_battr, strlen(s_battr)));
  nodspsym=string2symbol(makfromstr(s_nodsp, strlen(s_nodsp)));
  initscr();   /* should be delayed */
  start_color();
  nocbreak();
  nl();
  echo();
  noraw();
  conwin =newwin(0,0,0,0);
  *loc_console=makewin(conwin);
  console=*loc_console;
  tmpwin=newwin(0, 0, 0, 0);

  /* initialize colors */
  if (has_colors())
     {
     start_color();
     init_pair(0, COLOR_BLACK, COLOR_BLACK);
     init_pair(1, COLOR_RED, COLOR_BLACK);
     init_pair(2, COLOR_GREEN, COLOR_BLACK);
     init_pair(3, COLOR_MAGENTA, COLOR_BLACK);
     init_pair(4, COLOR_BLUE, COLOR_BLACK);
     init_pair(5, COLOR_CYAN, COLOR_BLACK);
     init_pair(6, COLOR_YELLOW, COLOR_BLACK);
     init_pair(7, COLOR_WHITE, COLOR_BLACK);
     }
  init_iprocs(subr0s, tc7_subr_0);
  init_iprocs(subr1s, tc7_subr_1);
  init_iprocs(subr2s, tc7_subr_2);
  init_iprocs(subr3s, tc7_subr_3);
  init_iprocs(lsubrs, tc7_lsubr);
  init_iprocs(subr2os, tc7_subr_2o);
#ifdef START_IN_WINMODE
  p_ewinmode(); 
/* should be delayed */
#endif
  /* initialize *window-list* */
  *loc_maxconsole=cons (MAKINUM(LINES), MAKINUM(COLS));
  *loc_winlist=cons(cons(console, 
			 cons(BOOL_F, 
			      cons(CAR(intern(CHARS(""),(sizet)0)),
				   cons(MAKINUM(TEXTATTR),
					cons(MAKINUM(BORDERATTR), 
					     EOL))))), 
		    *loc_winlist);
  add_feature("win");
#ifdef WINEXT
  add_feature("winext");
#endif
  add_final(finalxwin);
}
