/*
Copyright (C) 1995 John Leveling

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

This is Version 0.8.2 alpha of WinPCS, a
text-window extension for use with SCM.

The author can be reached at olga@informatik.uni-oldenburg.de or
J. Leveling, Sonnenstr. 32, 26123 Oldenburg, Germany
*/

#include "scm.h"
#include <curses.h>

/* define this if you want additional window procedures */
/* not supported by PC-Scheme */
#define WINEXT

/* define this if you want windows cleared by scrolling their contents */
/* it may be faster than using wclear() */
/* depends on scrolling capabilities */
#define CLEAR_BY_SCROLL 

/* set standard text and border attributes */
/* should be between 0 and 255 */
/* default : text 7 (white) , border 10 (green) */
#define TEXTATTR 7
#define BORDERATTR 10

/* macros for window size, window position and cursor position */
/* you may have to redefine these for other curses versions */
#define MAXX(window) (WIN(window)->_maxx)
#define MAXY(window) (WIN(window)->_maxy)
#define BEGX(window) (WIN(window)->_begx)
#define BEGY(window) (WIN(window)->_begy)
#define CURX(window) (WIN(window)->_curx)
#define CURY(window) (WIN(window)->_cury)

/* define this if you want immediate update of window contents */
#define IMMREFRESH 

/* define this if you want to start SCM in window-mode */
#define START_IN_WINMODE 

/* define this if you want to check status of window mode */
/* before executing a window related procedure */
/* undefining this should increase performance slightly */
#define CHECK_WINMODE

/* end of user-definable settings */



/* define WIN port type */
#define WIN(obj) ((WINDOW*)CDR(obj))
#define WINP(obj) (tc16_window==TYP16(obj))

#define WINDOWP(obj) (NIMP(obj) && WINP(obj))
#define SETCONSOLE if (win==consolesym) win=console
#define WINMODEP (winmodep==1)

static char e_invwinarg[]="\n; invalid window argument",
	e_invlineno[]="\n; invalid line number",
	e_invcolno[]="\n; invalid column number",
	e_invarg[]="\n; invalid argument",
	e_invmode[]="\n; not in window mode";

static char 

	s_mkwin[]=      "make-window",
	s_winclear[]=   "window-clear",    
	s_windelete[]=  "window-delete",  
	s_wingattr[]=   "window-get-attribute", 
	s_wingcurs[]=   "window-get-cursor",    
	s_wingpos[]=    "window-get-position",
	s_wingsize[]=   "window-get-size",
/*	s_winpopup[]=   "window-popup",  */
/*	s_windelpopup[]="window-popup-delete",  */ 
	s_winrestorec[]="window-restore-contents", 
	s_winsavec[]=   "window-save-contents", 
	s_winsattr[]=   "window-set-attribute!",
	s_winscurs[]=   "window-set-cursor!", 
	s_winspos[]=    "window-set-position!",
	s_winssize[]=   "window-set-size!",
	s_winp[]=       "window?",
	s_winscrlup[]=  "window-scroll-up",
	s_winscrldn[]=  "window-scroll-down",
#ifdef WINEXT
	s_newwin[]=     "new-window",
	s_winborder[]=  "window-border",
        s_windisplay[]= "window-display",
	s_winghelp[]=   "window-get-help",
	s_winrefresh[]= "window-refresh",
	s_wintitle[]=   "window-title",
#endif
	s_ewinmode[]=   "enter-window-mode",
	s_xwinmode[]=   "exit-window-mode";

#ifdef WINEXT
/* set this to the number of window procedures */
#define NWINPROCS 25

static char *helptext[NWINPROCS]={
"(enter-window-mode)\n\
   enters window mode.\n\
     \n\
     \n\
     \n",
"(exit-window-mode)\n\
   exits window mode.\n\
     \n\
     \n\
     \n",
"(make-window {label} {border?})\n\
   creates a window object.\n\
     label (string) specifies the window title\n\
     border? (boolean) specifies whether the window is to have a border\n\
     \n",
"(window? object)\n\
   returns #t if object is window and #f otherwise.\n\
     object (object) \n\
     \n\
     \n",
"(window-clear win)\n\
   clears a window and draws a border if one is specified.\n\
     win (window)\n\
     \n\
     \n",
"(window-delete win)\n\
   deletes a window object as well as its border.\n\
     win (window) the port associated with window is closed\n\
     \n\
     \n",
"(window-get-attribute win attrtype)\n\
   returns attributes.\n\
     win (window) \n\
     attrtype ('text-attributes or 'border-attributes) \n\
     \n",
"(window-get-cursor win)\n\
   returns cursor position in win as dotted pair.\n\
     win (window) (line . column)\n\
     positions are relative to the upper-lefthand corner of win\n\
     \n",
"(window-get-help)\n\
   displays this window.\n\
     to get further information about window procedures,\n\
     read the manual or PCScheme User's Guide\n\
     \n",
"(window-get-position win)\n\
   returns window position.\n\
     win (window) \n\
     \n\
     \n",
"(window-get-size win)\n\
   returns window size.\n\
     win (window) \n\
     \n\
     \n",
"(window-popup)\n\
   defines a popup-window.\n\
     \n\
     \n\
     \n",
"(window-popup-delete win)\n\
   deletes a popup-window.\n\
     win (window) \n\
     \n\
     \n",
"(window-restore-contents win contents)\n\
   restores window contents.\n\
     win (window) \n\
     contents (!!!) \n\
     \n",
"(window-save-contents win)\n\
   saves window contents.\n\
     win (window) \n\
     \n\
     \n",
"(window-scroll-down win {nlines})\n\
   scrolls contents of window down.\n\
     win (window) \n\
     nlines (integer) \n\
     \n",
"(window-scroll-up win {nlines})\n\
   scrolls contents of window up.\n\
     win (window) \n\
     nlines (integer) \n\
     \n",
"(window-set-attribute! win attrtype attr)\n\
   sets window attributes.\n\
     win (window) \n\
     attrtype ('text-attributes or 'border-attributes) \n\
     attr (integer) \n",
"(window-set-cursor! win line col)\n\
   sets cursor position.\n\
     win (window) \n\
     line, col (integers) \n\
     \n",
"(window-set-position! win ul-line ul-col)\n\
   sets window position.\n\
     win (window) \n\
     ul-line, ul-col (integers) \n\
     \n",
"(window-set-size! win nlines ncols)\n\
   sets window size.\n\
     win (window) \n\
     nlines, ncols (integers)\n\
     \n",
"(new-window title {{minlines} {mincols}} {'NO-DISPLAY})\n\
   lets you interactively place a window.\n\
     title (string) \n\
     minlines, mincols (integer) \n\
     'no-display (symbol) \n",
"(window-border win borderp)\n\
   specifies the border flag for window.\n\
     win (window) \n\
     borderp (boolean) \n\
     \n",
"(window-refresh win)\n\
   refreshes window contents.\n\
     win (window) \n\
     \n\
     \n",
"(window-display win line col string)\n\
   displays string in window at location (line, col).\n\
     win (window) \n\
     line, col (integers) and the string is displayed\n\
     string (string) \n"
};
#endif

SCM *loc_console=0;
SCM *loc_winlist=0;
SCM *loc_maxconsole=0;
SCM winmodep=0;
SCM console=0;
SCM consolesym=0;
SCM tattrsym=0;
SCM battrsym=0;
SCM nodspsym=0;
WINDOW *tmpwin;


/* The following were taken from Aubrey Jaffer's crs.c
and changed accordingly */

int pfreewindow(WINDOW *win)
{
  if (win==stdscr) return 0;
  delwin(win);
  return 0;
}

int pprinwindow(SCM exp, SCM port, int writing)
{
  prinport(exp,port,"window");
  return !0;
}

int pbwaddch(int c, WINDOW *win) 
{
  waddch(win,c);
#ifdef IMMREFRESH
  wrefresh(win); 
/* added */
#endif
  return c;
}

int pgetch(WINDOW *win)   /* added */
{
  int c;   /* wmove ?? */
#ifdef IMMREFRESH
  wrefresh(win);  
/* wrong order ?? */
#endif
  c=wgetch(win);
  return c;
}

int pbwaddstr(char *str, WINDOW *win) 
{
  waddstr(win,str);
#ifdef IMMREFRESH
  wrefresh(win); 
/* added */
#endif
  return 0;
}

sizet pbwwrite(char *str, sizet siz, sizet num,WINDOW *win)
{
  sizet i = 0, prod = siz*num;
  for (;i < prod;i++) waddch(win,str[i]);
#ifdef IMMREFRESH
  wrefresh(win);  
/* added */
#endif
  return num;
}

int tc16_window;
static ptobfuns winptob = {
  mark0,
  pfreewindow,
  pprinwindow,
  equal0,
  pbwaddch,
  pbwaddstr,
  pbwwrite,
  wrefresh,
  pgetch,
  pfreewindow};

/* return attributes from *window-list* */
SCM wingattr(SCM win)
{
  SCM wins=CDR(loc_winlist);
  while (CAR(CAR(wins))!=win)
	{wins=CDR(wins);}
  return CAR(wins);
}

/* create a new window */
SCM makewin(WINDOW *win)
{
  SCM z;
  if (NULL==win) return BOOL_F;
  NEWCELL(z);
  DEFER_INTS;
  SETCHARS(z, win);
  CAR(z) = tc16_window | OPN | RDNG | WRTNG;
  ALLOW_INTS;
  return z;
}

/* map integer to attributes */
chtype attributes(SCM attrs)
{
  chtype cattr=0;
  attrs=INUM(attrs);
  if (attrs>=128)
     {attrs-=128;cattr|=A_ALTCHARSET;} 
  if (attrs>=64)
     {attrs-=64;cattr|=A_BLINK;}
  if (attrs>=32)
     {attrs-=32;cattr|=A_UNDERLINE;}
  if (attrs>=16)
     {attrs-=16;cattr|=A_REVERSE;}
  if (attrs>=8)
     {attrs-=8;cattr|=A_BOLD;}
  if (has_colors())
     cattr|=COLOR_PAIR(attrs);
  return cattr;
}

/* draw a border just outside of window */
void borderwin(SCM win, int drawp)
{
  SCM wattr,battr,bf,title;
  int begy, begx, endy, endx, i=0;
  chtype ul=ACS_ULCORNER, ur=ACS_URCORNER, bl=ACS_LLCORNER, br=ACS_LRCORNER;
  chtype u=ACS_HLINE,b=ACS_HLINE,l=ACS_VLINE,r=ACS_VLINE,sp=' ';
  wattr=wingattr(win);
  bf=CAR(CDR(wattr));
  title=CAR(CDR(CDR(wattr)));
  battr=(CAR(CDR(CDR(CDR(CDR(wattr))))));
  begy=(BEGY(win)-1);
  begx=(BEGX(win)-1);
  endy=begy+MAXY(win)+2;
  endx=begx+MAXX(win)+2;

  if (begy<0)		        /* top not visible */
     {ul=ur=ACS_VLINE;u=sp;}
  if (endy>=LINES)	        /* bottom not visible */
     {bl=br=ACS_VLINE;b=sp;}
  if (begx<0)                   /* left not visible */
     {ul=bl=ACS_HLINE;l=sp;}
  if (endx>=COLS)		/* right not visible */
     {ur=br=ACS_HLINE;r=sp;}
  if (begy<0 && begx<0)                 /* upper left not visible */
     {bl=ACS_HLINE;ur=ACS_VLINE;ul=sp;}
  if (begy<0 && endx>=COLS)		/* upper right not visible */
     {br=ACS_HLINE;ul=ACS_VLINE;ur=sp;}
  if (endy>=LINES && begx<0)		/* bottom left not visible */
     {br=ACS_VLINE;ul=ACS_HLINE;bl=sp;}
  if (endy>=LINES && endx>=COLS)	/* bottom right not visible */
     {ur=ACS_HLINE;bl=ACS_VLINE;br=sp;}

  if (begy<0)      begy=0;		/* map to valid coordinates */
  if (begy>=LINES) begy=LINES-1; 
  if (begx<0)      begx=0;
  if (begx>=COLS)  begx=COLS-1;

  if (endy>=LINES) endy=LINES-1;
  if (endx>=COLS)  endx=COLS-1;

  tmpwin->_maxy=endy-begy;
  tmpwin->_maxx=endx-begx;
  tmpwin->_begy=begy;
  tmpwin->_begx=begx;

  if (drawp && bf==BOOL_T)
     {
     wattrset(tmpwin, attributes(battr));
     wborder(tmpwin, l, r, u, b, ul, ur, bl, br);
     }
  if (drawp)
     {
     tmpwin->_cury=0;
     tmpwin->_curx=1;
     while ((CHARS(title)[i]!='\0') &&   /* display whole string */
	    (i+1< endx-begx) &&          /* unlesss window too small */
	    (begx+i+1 <COLS))            /* or too far right */
/*	    (begy >0)) */
	{waddch(tmpwin, CHARS(title)[i]);
		i++;}
     }
  if (!drawp)
     wborder(tmpwin, sp, sp, sp, sp, sp, sp, sp, sp);
  wrefresh(tmpwin);
  return;
}


/* called by SCM when exiting */
void finalxwin(void)
{
  if (winmodep==0) return;
  winmodep=0;
  if IMP(*loc_console) return;
  wclear(WIN(console));
  wrefresh(WIN(console));
  cur_errp=def_errp;
  cur_inp=def_inp;
  cur_outp=def_outp;
  endwin();
}
