;;;; "winbench.scm"
;;; Copyright (C) 1995, J. Leveling
;;; See the file `COPYING' for terms applying to this program.

; 'Benchmark' for some procedures

(require 'win)
(require 'random)

(enter-window-mode)
(window-clear console)

; set up some windows
(define win1 (make-window "win1" #t))
(window-set-size! win1 4 4)
(window-set-position! win1 1 10)
(window-clear win1)

(define win2 (make-window "win2" #t))
(window-set-size! win2 5 5)
(window-set-position! win2 1 16)
(window-clear win2)

(define win3 (make-window "win3" #t))
(window-set-size! win3 6 6)
(window-set-position! win3 1 23)
(window-clear win3)

(define win4 (make-window "win4" #t))
(window-set-size! win4 7 7)
(window-set-position! win4 1 31)
(window-clear win4)

(define win5 (make-window "win5" #t))
(window-set-size! win5 8 8)
(window-set-position! win5 1 40)
(window-clear win5)

(define win6 (make-window "win6" #t))
(window-set-size! win6 9 9)
(window-set-position! win6 1 50)
(window-clear win6)

(define win7 (make-window "win7" #t))
(window-set-size! win7 10 10)
(window-set-position! win7 1 61)
(window-clear win7)

(define win8 (make-window "win8" #t))
(window-set-size! win8 4 7)
(window-set-position! win8 3 1)
(window-clear win8)

(define win9 (make-window "win9" #t))
(window-set-size! win9 2 28)
(window-set-position! win9 9 1)
(window-clear win9)

(define win10 (make-window "win10" #t))
(window-set-size! win10 6 65)
(window-set-position! win10 4 4)
(window-clear win10)

; list of windows
(define winlist
  (list win1 win2 win3 win4 win5 win6 win7 win8 win9 win10))

; choose a window
(define (rwin)
  (list-ref winlist (random 10)))
  
; set up output window
(define out (make-window "results" #t))
(window-set-size! out 11 70)
(window-set-position! out 13 1)
(window-clear out)	

; 'benchmark' routine
(define (wbench procedure counter)
  (define (bench c)
    (if (= c 0)
	(get-internal-run-time)
	(begin
	 (eval procedure)
	 (bench (- c 1)))))
  (display "Testing : " out)
  (display procedure out)
  (display " " out)
  (display counter out)
  (display " " out)
  (display	
   (round (* 100 	
	     (/ (abs (- (get-internal-run-time) (bench counter)))
		internal-time-units-per-second))) out)
  (newline out))	

(define (winbench)

  ; benchmark for output primitives
  (wbench '(newline (rwin)) 1000)

  (wbench '(display "0" (rwin)) 1000)
  (wbench '(display "0123456789" (rwin)) 1000)
  (wbench '(write "0" (rwin)) 1000)
  (wbench '(write "0123456789" (rwin)) 1000)
  (wbench '(write "0" (rwin)) 1000)
  (wbench '(write-line "0123456789" (rwin)) 1000) 

  ; benchmark for scrolling
  (wbench '(window-scroll-up (rwin)) 1000)
  (wbench '(window-scroll-down (rwin)) 1000)

  ; benchmark for get primitives
  (wbench '(window-get-cursor (rwin)) 1000) 
  (wbench '(window-get-position (rwin)) 1000) 
  (wbench '(window-get-size (rwin)) 1000) 
  (wbench '(window-get-attribute (rwin) 'text-attributes) 1000) 

  (wbench '(window-save-contents (rwin)) 1000)

  ; benchmark for clearing windows
  (wbench '(window-clear (rwin)) 100)

  ; test your own
  ; ...

  )				

; start 'benchmark'
(winbench)
