/*
Copyright (C) 1995 John Leveling

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

This is Version 0.8.2 alpha of WinPCS, a
text-window extension for use with SCM.

The author can be reached at olga@informatik.uni-oldenburg.de or
J. Leveling, Sonnenstr. 32, 26123 Oldenburg, Germany
*/

/* new-window */
/* mark corner */
/* still behaves different to PCS newwin */
SCM p_newwin(SCM args)
{
  int nargs=ilength(args), display=1;
  chtype key=0;
  SCM win, label, minlines=1, mincols=1, winattrs;
  WINDOW *nwin;
  ASSERT (WINMODEP, args, e_invmode, s_newwin);
  ASSERT ((nargs==1) || (nargs==3) || (nargs==4), args, WNA, s_newwin);
  ASSERT ((CAR(args)==BOOL_F) || STRINGP(CAR(args)),
	  CAR(args), e_invarg, s_newwin);
  if (nargs>=3)
     {
     minlines=CAR(CDR(args));            /* minlines >0 & < LINES */
     ASSERT(INUMP(minlines)||minlines==BOOL_F, 
         minlines, e_invarg, s_newwin);
     mincols=CAR(CDR(CDR(args)));
     ASSERT(INUMP(mincols)||mincols==BOOL_F, 
         mincols, e_invarg, s_newwin);
     }
  if (nargs==4)
     {   /* no-display */
     ASSERT((nodspsym==CAR(CDR(CDR(CDR(args))))), 
	    CAR(CDR(CDR(CDR(args)))), e_invarg, s_newwin);
     display=0;
     }
  nwin=newwin(0,0,0,0);
  win=makewin(nwin);
  keypad(WIN(win), TRUE);
  idlok(WIN(win), TRUE);
  scrollok(WIN(win), TRUE);
  wsetscrreg(WIN(win), 0, MAXY(win));
  label=CAR(args);
  label=CAR(intern(CHARS(label),(sizet)LENGTH(label)));
  winattrs=cons(win,				   /* port */
	        cons(BOOL_T,			   /* window-bordered? */
		     cons(label,		   /* window's title */
			  cons(MAKINUM(TEXTATTR),  /* attributes */
			       cons(MAKINUM(BORDERATTR), /* borderattrs */
				    EOL)))));
  *loc_winlist=cons(winattrs, *loc_winlist);
  BEGY(win)=INUM(1);
  BEGX(win)=INUM(1);
  MAXY(win)=INUM(minlines);
  MAXX(win)=INUM(mincols);

  while (key!='\n')
	{
	if ((key==KEY_UP)    && (BEGY(win)>0))     
           {borderwin(win, 0);BEGY(win)--;}
	if ((key==KEY_DOWN)  && (BEGY(win)<LINES-1)) 
	   {borderwin(win, 0);BEGY(win)++;}
	if ((key==KEY_LEFT)  && (BEGX(win)>0))     
	   {borderwin(win, 0);BEGX(win)--;}
	if ((key==KEY_RIGHT) && (BEGX(win)<COLS-1))  
	   {borderwin(win, 0);BEGX(win)++;}
	borderwin(win, display); 
	CURX(win)=0;
	CURY(win)=0;
	wrefresh(WIN(win));
	key=wgetch(WIN(win));
	}
  key=0;
  while (key!='\n')
	{
	if ((key==KEY_UP)    && (MAXY(win)>0))     
           {borderwin(win, 0);MAXY(win)--;}
	if ((key==KEY_DOWN)  && (MAXY(win)+BEGY(win)<=LINES)) 
	   {borderwin(win, 0);MAXY(win)++;}
	if ((key==KEY_LEFT)  && (MAXX(win)>0))     
	   {borderwin(win, 0);MAXX(win)--;}
	if ((key==KEY_RIGHT) && (MAXX(win)+BEGX(win)<=COLS))  
	   {borderwin(win, 0);MAXX(win)++;}
	borderwin(win, display); 
	CURX(win)=0;
	CURY(win)=0;
	wrefresh(WIN(win));
	key=wgetch(WIN(win));
	}
  if (display==1)
     {
/* wclear(WIN(win)); */
     wrefresh(WIN(win));
     }
  return win;
}

/* window-border */
SCM p_winborder (SCM win, SCM borderp)
{
  SCM attrs;
  ASSERT (WINMODEP, win, e_invmode, s_winborder);
  SETCONSOLE;
  ASSERT(WINDOWP(win), win, e_invwinarg, s_winborder);
  ASSERT(UNBNDP(borderp) || (borderp==BOOL_T) || (borderp==BOOL_F), borderp, 
	 e_invarg, s_winborder);
  attrs=wingattr(win);
  if (UNBNDP(borderp))
     return (CAR(CDR(attrs))==BOOL_T) ? BOOL_T : BOOL_F;
  CAR(CDR(attrs))=borderp;
  if (borderp==BOOL_F) 
	borderwin(win, 0);
     else 
	borderwin(win, 1);
  wrefresh(WIN(win));
  return UNSPECIFIED;
}

SCM p_windisplay(SCM args)
{
  int nargs=ilength(args);
  short lines, columns;
  SCM win, line, column, string;
  ASSERT (nargs==4, args, WNA, s_windisplay);
  win=CAR(args);
  line=CAR(CDR(args));
  column=CAR(CDR(CDR(args)));
  string=CAR(CDR(CDR(CDR(args))));
  ASSERT (WINMODEP, win, e_invmode, s_windisplay);
  SETCONSOLE;
  ASSERT(WINDOWP(win), win, e_invarg, s_windisplay);
  ASSERT(INUMP(line)||line==BOOL_F, 
         line, e_invarg, s_windisplay);
  ASSERT(INUMP(column)||column==BOOL_F, 
         column, e_invarg, s_windisplay);
  ASSERT (STRINGP(string),
	  string, e_invarg, s_windisplay);
  if (line!=BOOL_F)
     {
     lines=MAXY(win);
     if (INUMP(line))
	line=INUM(line);
	else line=REALPART(line);
     ASSERT((line>=0)&&(line<=lines), 
	    MAKINUM(line), e_invlineno, s_windisplay);
     }
  if (column!=BOOL_F)
     {
     columns=MAXX(win);
     if (INUMP(column))
	column=INUM(column);
	else column=REALPART(column);
     ASSERT((column>=0)&&(column<=columns), 
	    MAKINUM(column), e_invcolno, s_windisplay);
     }
  if (line!=BOOL_F) CURY(win)=line;
  if (column!=BOOL_F) CURX(win)=column;
  string=CAR(intern(CHARS(string),(sizet)LENGTH(string)));
  waddstr(WIN(win), CHARS(string));
return UNSPECIFIED;
}

/* window-get-help */
SCM p_winghelp(void)
{
/* wta 0 */
  short i,c=0;
  chtype key=0;
  ASSERT (WINMODEP, MAKINUM(0), e_invmode, s_winghelp);
  keypad(tmpwin, TRUE);
  tmpwin->_maxx=COLS;
  tmpwin->_maxy=LINES;
  tmpwin->_begy=0;
  tmpwin->_begx=0;
  wclear(tmpwin);
  waddstr(tmpwin, "Format               Description                        \n");
  waddstr(tmpwin, "========================================================\n");
  while ((key!=KEY_LEFT) && (key!=KEY_RIGHT) && (key!='\n'))
        {
        tmpwin->_cury=2;
        tmpwin->_curx=0;
        for (i=c;i<c+3;i++)    /* display 3 simultaneously */
	    {
	    waddstr(tmpwin, helptext[i]);
            }
        wrefresh(tmpwin);
        key=wgetch(tmpwin);
        if (key==KEY_UP && c>0) c--;
        if (key==KEY_DOWN && c+3<NWINPROCS) c++;
	}
  wclear(tmpwin);
  wrefresh(tmpwin);
  return UNSPECIFIED;
}

SCM p_winrefresh(SCM win)
{
  ASSERT (WINMODEP, win, e_invmode, s_winrefresh);
  SETCONSOLE;
  ASSERT(WINDOWP(win), win, e_invwinarg, s_winrefresh);
  borderwin(win, 1);
  wrefresh(WIN(win));
  return UNSPECIFIED;
}

/* window-title */
SCM p_wintitle (SCM win, SCM title)
{
  SCM attrs;
  ASSERT (WINMODEP, win, e_invmode, s_wintitle);
  SETCONSOLE;
  ASSERT(WINDOWP(win), win, e_invwinarg, s_wintitle);
  ASSERT(UNBNDP(title) || STRINGP(title), title,
	 e_invarg, s_wintitle);
  attrs=wingattr(win);
  if (UNBNDP(title))
     {
     title=(CAR(CDR(CDR(attrs))));
     return makfromstr(CHARS(title), (sizet)LENGTH(title));
     }
  title=CAR(intern(CHARS(title),(sizet)LENGTH(title)));
  CAR(CDR(CDR(attrs)))=title;
  wrefresh(WIN(win));
  return UNSPECIFIED;
}

